from cassio.table.tables import PlainCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredMetadataCassandraTable  # noqa: F401
from cassio.table.tables import MetadataCassandraTable  # noqa: F401
from cassio.table.tables import VectorCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredVectorCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredMetadataVectorCassandraTable  # noqa: F401
from cassio.table.tables import MetadataVectorCassandraTable  # noqa: F401
from cassio.table.tables import ElasticCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredElasticCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredElasticMetadataCassandraTable  # noqa: F401
from cassio.table.tables import ElasticMetadataCassandraTable  # noqa: F401
from cassio.table.tables import ElasticVectorCassandraTable  # noqa: F401
from cassio.table.tables import ClusteredElasticVectorCassandraTable  # noqa: F401
from cassio.table.tables import (
    ClusteredElasticMetadataVectorCassandraTable,  # noqa: F401
)
from cassio.table.tables import ElasticMetadataVectorCassandraTable  # noqa: F401



"""
An extractor able to resolve single-row lookups from Cassandra tables in
a keyspace, with a fair amount of metadata inspection.
"""

from functools import reduce
from typing import List

from cassandra.query import SimpleStatement


RETRIEVE_ONE_ROW_CQL_TEMPLATE = 'SELECT * FROM {keyspace}.{table} WHERE {whereClause} LIMIT 1'


def _table_primary_key_columns(session, keyspace, table) -> List[str]:
    table = session.cluster.metadata.keyspaces[keyspace].tables[table]
    return [
        col.name for col in table.partition_key
    ] + [
        col.name for col in table.clustering_key
    ]


class CassandraExtractor:

    def __init__(self, session, keyspace, field_mapper, literal_nones):
        self.session = session
        self.keyspace = keyspace
        self.field_mapper = field_mapper
        self.literal_nones = literal_nones  # TODO: handle much better
        # derived fields
        self.tables_needed = {fmv[0] for fmv in field_mapper.values()}
        self.primary_key_map = {
            table: _table_primary_key_columns(self.session, self.keyspace, table)
            for table in self.tables_needed
        }
        # all primary-key values needed across tables
        self.requiredParameters = list(reduce(lambda accum, nw: accum | set(nw), self.primary_key_map.values(), set()))

        # TODOs:
        #   move this getter creation someplace else
        #   query a table only once (grouping required variables by source table,
        #   selecting only those unless function passed)
        def _getter(**kwargs):
            def _retrieve_field(_table2, _key_columns, _column_or_extractor, _key_value_map):
                selector = SimpleStatement(RETRIEVE_ONE_ROW_CQL_TEMPLATE.format(
                    keyspace=keyspace,
                    table=_table2,
                    whereClause=' AND '.join(
                        f'{kc} = %s'
                        for kc in _key_columns
                    ),
                ))
                values = tuple([
                    _key_value_map[kc]
                    for kc in _key_columns
                ])
                row = session.execute(selector, values).one()
                if row:
                    if callable(_column_or_extractor):
                        return _column_or_extractor(row)
                    else:
                        return getattr(row, _column_or_extractor)
                else:
                    if literal_nones:
                        return None
                    else:
                        raise ValueError('No data found for %s from %s.%s' % (
                            str(_column_or_extractor),
                            keyspace,
                            _table2,
                        ))

            return {
                field: _retrieve_field(table, self.primary_key_map[table], columnOrExtractor, kwargs)
                for field, (table, columnOrExtractor) in field_mapper.items()
            }

        self.getter = _getter

    def __call__(self, **kwargs):
        return self.getter(**kwargs)



from typing import Union
from enum import Enum

from cassandra.query import SimpleStatement, PreparedStatement  # type: ignore


class CQLOpType(Enum):
    SCHEMA = 1
    WRITE = 2
    READ = 3


CREATE_TABLE_CQL_TEMPLATE = """CREATE TABLE IF NOT EXISTS {{table_fqname}} ({columns_spec} PRIMARY KEY {primkey_spec}) {clustering_spec};"""  # noqa: E501

TRUNCATE_TABLE_CQL_TEMPLATE = """TRUNCATE TABLE {{table_fqname}};"""

DELETE_CQL_TEMPLATE = """DELETE FROM {{table_fqname}} {where_clause};"""

SELECT_CQL_TEMPLATE = (
    """SELECT {columns_desc} FROM {{table_fqname}} {where_clause} {limit_clause};"""
)

INSERT_ROW_CQL_TEMPLATE = """INSERT INTO {{table_fqname}} ({columns_desc}) VALUES ({value_placeholders}) {ttl_spec} ;"""  # noqa: E501

CREATE_INDEX_CQL_TEMPLATE = """CREATE CUSTOM INDEX IF NOT EXISTS {index_name}_{{table_name}} ON {{table_fqname}} ({index_column}) USING 'org.apache.cassandra.index.sai.StorageAttachedIndex';"""  # noqa: E501

CREATE_KEYS_INDEX_CQL_TEMPLATE = """CREATE CUSTOM INDEX IF NOT EXISTS {index_name}_{{table_name}} ON {{table_fqname}} (KEYS({index_column})) USING 'org.apache.cassandra.index.sai.StorageAttachedIndex';"""  # noqa: E501

CREATE_ENTRIES_INDEX_CQL_TEMPLATE = """CREATE CUSTOM INDEX IF NOT EXISTS {index_name}_{{table_name}} ON {{table_fqname}} (ENTRIES({index_column})) USING 'org.apache.cassandra.index.sai.StorageAttachedIndex';"""  # noqa: E501

SELECT_ANN_CQL_TEMPLATE = """SELECT {columns_desc} FROM {{table_fqname}} {where_clause} ORDER BY {vector_column} ANN OF %s {limit_clause};"""  # noqa: E501

CQLStatementType = Union[str, SimpleStatement, PreparedStatement]

# Mock DB session
class MockDBSession:
    def __init__(self, verbose=False):
        self.verbose = verbose
        self.statements = []

    @staticmethod
    def getStatementBody(statement: CQLStatementType) -> str:
        if isinstance(statement, str):
            _statement = statement
        elif isinstance(statement, SimpleStatement):
            _statement = statement.query_string
        elif isinstance(statement, PreparedStatement):
            _statement = statement.query_string
        else:
            raise ValueError()
        return _statement

    @staticmethod
    def normalizeCQLStatement(statement: CQLStatementType) -> str:
        _statement = MockDBSession.getStatementBody(statement)
        _s = (
            _statement.replace(";", " ")
            .replace("%s", " %s ")
            .replace("?", " ? ")
            .replace("=", " = ")
            .replace(")", " ) ")
            .replace("(", " ( ")
            .replace("\n", " ")
        )
        return " ".join(
            tok.lower() for tok in (_t.strip() for _t in _s.split(" ") if _t.strip())
        )

    @staticmethod
    def prepare(statement):
        # A very unusable 'prepared statement' just for tracing/debugging:
        return PreparedStatement(None, 0, 0, statement, "keyspace", None, None, None)

    def execute(self, statement, arguments=tuple()):
        if self.verbose:
            #
            st_body = self.getStatementBody(statement)
            if isinstance(statement, str):
                st_type = "STR"
                placeholder_count = st_body.count("%s")
                assert "?" not in st_body
            elif isinstance(statement, SimpleStatement):
                st_type = "SIM"
                placeholder_count = st_body.count("%s")
                assert "?" not in st_body
            elif isinstance(statement, PreparedStatement):
                st_type = "PRE"
                placeholder_count = st_body.count("?")
                assert "%s" not in st_body
            #
            assert placeholder_count == len(arguments)
            #
            print(f"CQL_EXECUTE [{st_type}]:")
            print(f"    {st_body}")
            if arguments:
                print(f"    {str(arguments)}")
        self.statements.append((statement, arguments))
        return []

    def last_raw(self, n):
        if n <= 0:
            return []
        else:
            return self.statements[-n:]

    def last(self, n):
        return [
            (
                self.normalizeCQLStatement(stmt),
                data,
            )
            for stmt, data in self.last_raw(n)
        ]

    def assert_last_equal(self, expected_statements):
        # used for testing
        last_executed = self.last(len(expected_statements))
        assert len(last_executed) == len(expected_statements)
        for s_exe, s_expe in zip(last_executed, expected_statements):
            assert s_exe[1] == s_expe[1], f"EXE#{str(s_exe[1])}# != EXPE#{s_expe[1]}#"
            exe_cql = self.normalizeCQLStatement(s_exe[0])
            expe_cql = self.normalizeCQLStatement(s_expe[0])
            assert exe_cql == expe_cql, f"EXE#{exe_cql}# != EXPE#{expe_cql}#"



from cassio.table.base_table import BaseTable
from cassio.table.mixins import (
    ClusteredMixin,
    MetadataMixin,
    VectorMixin,
    ElasticKeyMixin,
    #
    TypeNormalizerMixin,
)


class PlainCassandraTable(TypeNormalizerMixin, BaseTable):
    pass


class ClusteredCassandraTable(TypeNormalizerMixin, ClusteredMixin, BaseTable):
    clustered = True
    pass


class ClusteredMetadataCassandraTable(
    TypeNormalizerMixin, MetadataMixin, ClusteredMixin, BaseTable
):
    clustered = True
    pass


class MetadataCassandraTable(TypeNormalizerMixin, MetadataMixin, BaseTable):
    pass


class VectorCassandraTable(TypeNormalizerMixin, VectorMixin, BaseTable):
    pass


class ClusteredVectorCassandraTable(
    TypeNormalizerMixin, VectorMixin, ClusteredMixin, BaseTable
):
    clustered = True
    pass


class ClusteredMetadataVectorCassandraTable(
    TypeNormalizerMixin, MetadataMixin, ClusteredMixin, VectorMixin, BaseTable
):
    clustered = True
    pass


class MetadataVectorCassandraTable(
    TypeNormalizerMixin, MetadataMixin, VectorMixin, BaseTable
):
    pass


class ElasticCassandraTable(TypeNormalizerMixin, ElasticKeyMixin, BaseTable):
    elastic = True
    pass


class ClusteredElasticCassandraTable(
    TypeNormalizerMixin, ClusteredMixin, ElasticKeyMixin, BaseTable
):
    clustered = True
    elastic = True
    pass


class ClusteredElasticMetadataCassandraTable(
    TypeNormalizerMixin, MetadataMixin, ElasticKeyMixin, ClusteredMixin, BaseTable
):
    clustered = True
    elastic = True
    pass


class ElasticMetadataCassandraTable(
    TypeNormalizerMixin, MetadataMixin, ElasticKeyMixin, BaseTable
):
    elastic = True
    pass


class ElasticVectorCassandraTable(
    TypeNormalizerMixin, VectorMixin, ElasticKeyMixin, BaseTable
):
    elastic = True
    pass


class ClusteredElasticVectorCassandraTable(
    TypeNormalizerMixin, ClusteredMixin, ElasticKeyMixin, VectorMixin, BaseTable
):
    clustered = True
    elastic = True
    pass


class ClusteredElasticMetadataVectorCassandraTable(
    TypeNormalizerMixin,
    MetadataMixin,
    ElasticKeyMixin,
    ClusteredMixin,
    VectorMixin,
    BaseTable,
):
    clustered = True
    elastic = True
    pass


class ElasticMetadataVectorCassandraTable(
    MetadataMixin, ElasticKeyMixin, VectorMixin, BaseTable
):
    elastic = True
    pass



"""
management of "history" of stored blobs, grouped
by some 'session id'. Overwrites are not supported by design.
"""

import uuid
from warnings import warn
from typing import Any, Dict, Iterable, Optional

from cassandra.cluster import Session  # type: ignore

from cassio.table.tables import ClusteredCassandraTable


class StoredBlobHistory:
    """
    This class is a rewriting of the StoredBlobHistory created for use in LangChain
    integration, this time relying on the class-table-hierarchy (cassio.table.*).

    It mostly provides a translation layer between parameters and key names,
    using a clustered table class internally.

    Additional kwargs, for use in this new table class, are passed as they are
    in order to enable their usage already before adapting the LangChain
    integration code.
    """

    DEPRECATION_MESSAGE = (
        "Class `StoredBlobHistory` is a legacy construct and "
        "will be deprecated in future versions of CassIO."
    )

    def __init__(
        self, session: Session, keyspace: str, table_name: str, **kwargs: Dict[str, Any]
    ):
        #
        warn(self.DEPRECATION_MESSAGE, DeprecationWarning, stacklevel=2)
        # specifications are added such as the type of the row_id
        full_kwargs = {
            **kwargs,
            **{
                "primary_key_type": ["TEXT", "TIMEUUID"],
                # latest entries are returned first
                "ordering_in_partition": "DESC",
            },
        }
        self.table = ClusteredCassandraTable(
            session=session,
            keyspace=keyspace,
            table=table_name,
            **full_kwargs,
        )

    def store(self, session_id: str, blob: str, ttl_seconds: Optional[int]):
        this_row_id = uuid.uuid1()
        self.table.put(
            partition_id=session_id,
            row_id=this_row_id,
            body_blob=blob,
            ttl_seconds=ttl_seconds,
        )

    def retrieve(
        self, session_id: str, max_count: Optional[int] = None
    ) -> Iterable[str]:
        # The latest are returned, in chronological order
        return [
            row["body_blob"]
            for row in self.table.get_partition(
                partition_id=session_id,
                n=max_count,
            )
        ][::-1]

    def clear_session_id(self, session_id: str) -> None:
        self.table.delete_partition(session_id)
        return None



from typing import Any, List, Dict, Iterable, Optional, Set, Tuple, Union

from cassandra.query import SimpleStatement, PreparedStatement  # type: ignore
from cassandra.cluster import ResultSet  # type: ignore
from cassandra.cluster import ResponseFuture  # type: ignore

from cassio.table.table_types import (
    ColumnSpecType,
    RowType,
    SessionType,
    normalize_type_desc,
)
from cassio.table.cql import (
    CQLOpType,
    CREATE_TABLE_CQL_TEMPLATE,
    TRUNCATE_TABLE_CQL_TEMPLATE,
    DELETE_CQL_TEMPLATE,
    SELECT_CQL_TEMPLATE,
    INSERT_ROW_CQL_TEMPLATE,
)


class BaseTable:

    ordering_in_partition: Optional[str] = None

    def __init__(
        self,
        session: SessionType,
        keyspace: str,
        table: str,
        ttl_seconds: Optional[int] = None,
        row_id_type: Union[str, List[str]] = ["TEXT"],
        skip_provisioning=False,
    ) -> None:
        self.session = session
        self.keyspace = keyspace
        self.table = table
        self.ttl_seconds = ttl_seconds
        self.row_id_type = normalize_type_desc(row_id_type)
        self.skip_provisioning = skip_provisioning
        self._prepared_statements: Dict[str, PreparedStatement] = {}
        self.db_setup()

    def _schema_row_id(self) -> List[ColumnSpecType]:
        assert len(self.row_id_type) == 1
        return [
            ("row_id", self.row_id_type[0]),
        ]

    def _schema_pk(self) -> List[ColumnSpecType]:
        return self._schema_row_id()

    def _schema_cc(self) -> List[ColumnSpecType]:
        return []

    def _schema_da(self) -> List[ColumnSpecType]:
        return [
            ("body_blob", "TEXT"),
        ]

    def _schema(self) -> Dict[str, List[ColumnSpecType]]:
        return {
            "pk": self._schema_pk(),
            "cc": self._schema_cc(),
            "da": self._schema_da(),
        }

    def _schema_primary_key(self) -> List[ColumnSpecType]:
        return self._schema_pk() + self._schema_cc()

    def _schema_collist(self) -> List[ColumnSpecType]:
        full_list = self._schema_da() + self._schema_cc() + self._schema_pk()
        return full_list

    def _schema_colnameset(self) -> Set[str]:
        full_list = self._schema_collist()
        full_set = set(col for col, _ in full_list)
        assert len(full_list) == len(full_set)
        return full_set

    def _desc_table(self) -> str:
        columns = self._schema()
        col_str = (
            "[("
            + ", ".join("%s(%s)" % colspec for colspec in columns["pk"])
            + ") "
            + ", ".join("%s(%s)" % colspec for colspec in columns["cc"])
            + "] "
            + ", ".join("%s(%s)" % colspec for colspec in columns["da"])
        )
        return col_str

    def _extract_where_clause_blocks(
        self, args_dict: Any
    ) -> Tuple[Any, List[str], Tuple[Any, ...]]:
        # Removes some of the passed kwargs and returns the remaining,
        # plus the pieces for a WHERE
        _allowed_colspecs = self._schema_collist()
        passed_columns = sorted(
            [col for col, _ in _allowed_colspecs if col in args_dict]
        )
        residual_args = {k: v for k, v in args_dict.items() if k not in passed_columns}
        where_clause_blocks = [f"{col} = %s" for col in passed_columns]
        where_clause_vals = tuple([args_dict[col] for col in passed_columns])
        return (
            residual_args,
            where_clause_blocks,
            where_clause_vals,
        )

    def _normalize_kwargs(self, args_dict: Dict[str, Any]) -> Dict[str, Any]:
        return args_dict

    def _normalize_row(self, raw_row: Any) -> Dict[str, Any]:
        if isinstance(raw_row, dict):
            dict_row = raw_row
        else:
            dict_row = raw_row._asdict()
        #
        return dict_row

    def _delete(self, is_async: bool, **kwargs: Any) -> Union[None, ResponseFuture]:
        n_kwargs = self._normalize_kwargs(kwargs)
        (
            rest_kwargs,
            where_clause_blocks,
            delete_cql_vals,
        ) = self._extract_where_clause_blocks(n_kwargs)
        assert rest_kwargs == {}
        where_clause = "WHERE " + " AND ".join(where_clause_blocks)
        delete_cql = DELETE_CQL_TEMPLATE.format(
            where_clause=where_clause,
        )
        if is_async:
            return self.execute_cql_async(
                delete_cql, args=delete_cql_vals, op_type=CQLOpType.WRITE
            )
        else:
            self.execute_cql(delete_cql, args=delete_cql_vals, op_type=CQLOpType.WRITE)
            return None

    def delete(self, **kwargs: Any) -> None:
        self._delete(is_async=False, **kwargs)
        return None

    def delete_async(self, **kwargs: Any) -> ResponseFuture:
        return self._delete(is_async=True, **kwargs)

    def _clear(self, is_async: bool) -> Union[None, ResponseFuture]:
        truncate_table_cql = TRUNCATE_TABLE_CQL_TEMPLATE.format()
        if is_async:
            return self.execute_cql_async(
                truncate_table_cql, args=tuple(), op_type=CQLOpType.WRITE
            )
        else:
            self.execute_cql(truncate_table_cql, args=tuple(), op_type=CQLOpType.WRITE)
            return None

    def clear(self) -> None:
        self._clear(is_async=False)
        return None

    def clear_async(self) -> ResponseFuture:
        return self._clear(is_async=True)

    def _parse_select_core_params(
        self, **kwargs: Any
    ) -> Tuple[str, str, Tuple[Any, ...]]:
        n_kwargs = self._normalize_kwargs(kwargs)
        # TODO: work on a columns: Optional[List[str]] = None
        # (but with nuanced handling of the column-magic we have here)
        columns = None
        if columns is None:
            columns_desc = "*"
        else:
            # TODO: handle translations here?
            # columns_desc = ", ".join(columns)
            raise NotImplementedError("Column selection is not implemented.")
        #
        (
            rest_kwargs,
            where_clause_blocks,
            select_cql_vals,
        ) = self._extract_where_clause_blocks(n_kwargs)
        assert rest_kwargs == {}
        where_clause = "WHERE " + " AND ".join(where_clause_blocks)
        return columns_desc, where_clause, select_cql_vals

    def get(self, **kwargs: Any) -> Optional[RowType]:
        columns_desc, where_clause, get_cql_vals = self._parse_select_core_params(
            **kwargs
        )
        limit_clause = ""
        limit_cql_vals: List[Any] = []
        select_vals = tuple(list(get_cql_vals) + limit_cql_vals)
        #
        select_cql = SELECT_CQL_TEMPLATE.format(
            columns_desc=columns_desc,
            where_clause=where_clause,
            limit_clause=limit_clause,
        )
        # dancing around the result set (to comply with type checking):
        result_set = self.execute_cql(
            select_cql, args=select_vals, op_type=CQLOpType.READ
        )
        if isinstance(result_set, ResultSet):
            result = result_set.one()
        else:
            result = None
        #
        if result is None:
            return result
        else:
            return self._normalize_row(result)

    def get_async(self, **kwargs) -> ResponseFuture:
        raise NotImplementedError("Asynchronous reads are not supported.")

    def _put(self, is_async: bool, **kwargs: Any) -> Union[None, ResponseFuture]:
        n_kwargs = self._normalize_kwargs(kwargs)
        primary_key = self._schema_primary_key()
        assert set(col for col, _ in primary_key) - set(n_kwargs.keys()) == set()
        columns = [col for col, _ in self._schema_collist() if col in n_kwargs]
        columns_desc = ", ".join(columns)
        insert_cql_vals = [n_kwargs[col] for col in columns]
        value_placeholders = ", ".join("%s" for _ in columns)
        #
        ttl_seconds = (
            n_kwargs["ttl_seconds"] if "ttl_seconds" in n_kwargs else self.ttl_seconds
        )
        if ttl_seconds is not None:
            ttl_spec = "USING TTL %s"
            ttl_vals = [ttl_seconds]
        else:
            ttl_spec = ""
            ttl_vals = []
        #
        insert_cql_args = tuple(insert_cql_vals + ttl_vals)
        insert_cql = INSERT_ROW_CQL_TEMPLATE.format(
            columns_desc=columns_desc,
            value_placeholders=value_placeholders,
            ttl_spec=ttl_spec,
        )
        #
        if is_async:
            return self.execute_cql_async(
                insert_cql, args=insert_cql_args, op_type=CQLOpType.WRITE
            )
        else:
            self.execute_cql(insert_cql, args=insert_cql_args, op_type=CQLOpType.WRITE)
            return None

    def put(self, **kwargs: Any) -> None:
        self._put(is_async=False, **kwargs)
        return None

    def put_async(self, **kwargs: Any) -> ResponseFuture:
        return self._put(is_async=True, **kwargs)

    def db_setup(self) -> None:
        _schema = self._schema()
        column_specs = [
            f"{col_spec[0]} {col_spec[1]}"
            for _schema_grp in ["pk", "cc", "da"]
            for col_spec in _schema[_schema_grp]
        ]
        pk_spec = ", ".join(col for col, _ in _schema["pk"])
        cc_spec = ", ".join(col for col, _ in _schema["cc"])
        primkey_spec = f"( ( {pk_spec} ) {',' if _schema['cc'] else ''} {cc_spec} )"
        if _schema["cc"]:
            clu_core = ", ".join(
                f"{col} {self.ordering_in_partition}" for col, _ in _schema["cc"]
            )
            clustering_spec = f"WITH CLUSTERING ORDER BY ({clu_core})"
        else:
            clustering_spec = ""
        #
        create_table_cql = CREATE_TABLE_CQL_TEMPLATE.format(
            columns_spec=" ".join(f"  {cs}," for cs in column_specs),
            primkey_spec=primkey_spec,
            clustering_spec=clustering_spec,
        )
        self.execute_cql(create_table_cql, op_type=CQLOpType.SCHEMA)

    def _finalize_cql_semitemplate(self, cql_semitemplate: str) -> str:
        table_fqname = f"{self.keyspace}.{self.table}"
        table_name = self.table
        final_cql = cql_semitemplate.format(
            table_fqname=table_fqname, table_name=table_name
        )
        return final_cql

    def _obtain_prepared_statement(self, final_cql) -> PreparedStatement:
        # TODO: improve this placeholder handling
        _preparable_cql = final_cql.replace("%s", "?")
        # handle the cache of prepared statements
        if _preparable_cql not in self._prepared_statements:
            self._prepared_statements[_preparable_cql] = self.session.prepare(
                _preparable_cql
            )
        return self._prepared_statements[_preparable_cql]

    def execute_cql(
        self,
        cql_semitemplate: str,
        op_type: CQLOpType,
        args: Tuple[Any, ...] = tuple(),
    ) -> Iterable[RowType]:
        final_cql = self._finalize_cql_semitemplate(cql_semitemplate)
        #
        if op_type == CQLOpType.SCHEMA and self.skip_provisioning:
            # these operations are not executed for this instance:
            return []
        else:
            if op_type == CQLOpType.SCHEMA:
                # schema operations are not to be 'prepared'
                statement = SimpleStatement(final_cql)
            else:
                statement = self._obtain_prepared_statement(final_cql)
            #
            return self.session.execute(statement, args)

    def execute_cql_async(
        self,
        cql_semitemplate: str,
        op_type: CQLOpType,
        args: Tuple[Any, ...] = tuple(),
    ) -> ResponseFuture:
        final_cql = self._finalize_cql_semitemplate(cql_semitemplate)
        #
        if op_type == CQLOpType.SCHEMA:
            raise RuntimeError("Schema operations cannot be asynchronous")
        else:
            statement = self._obtain_prepared_statement(final_cql)
            #
            return self.session.execute_async(statement, args)



"""
handling of key-value storage on a Cassandra table.
One row per partition, serializes a multiple partition key into a string
"""

from warnings import warn
from typing import Any, Dict, List, Optional

from cassandra.cluster import Session  # type: ignore

from cassio.table.tables import ElasticCassandraTable


class KVCache:
    """
    This class is a rewriting of the KVCache created for use in LangChain
    integration, this time relying on the class-table-hierarchy (cassio.table.*).

    It mostly provides a translation layer between parameters and key names,
    using a clustered table class internally.

    Additional kwargs, for use in this new table class, are passed as they are
    in order to enable their usage already before adapting the LangChain
    integration code.
    """

    DEPRECATION_MESSAGE = (
        "Class `KVCache` is a legacy construct and "
        "will be deprecated in future versions of CassIO."
    )

    def __init__(self, session: Session, keyspace: str, table: str, keys: List[Any]):
        #
        warn(self.DEPRECATION_MESSAGE, DeprecationWarning, stacklevel=2)
        # for LangChain this is what we expect - no other uses are planned:
        assert all(isinstance(k, str) for k in keys)
        p_k_type = ["TEXT"] * len(keys)
        #
        self.table = ElasticCassandraTable(
            session,
            keyspace,
            table,
            keys=keys,
            primary_key_type=p_k_type,
        )

    def clear(self) -> None:
        self.table.clear()
        return None

    def put(
        self,
        key_dict: Dict[str, str],
        cache_value: str,
        ttl_seconds: Optional[int] = None,
    ) -> None:
        self.table.put(body_blob=cache_value, ttl_seconds=ttl_seconds, **key_dict)
        return None

    def get(self, key_dict) -> Optional[str]:
        entry = self.table.get(**key_dict)
        if entry is None:
            return None
        else:
            return entry["body_blob"]

    def delete(self, key_dict) -> None:
        """Will not complain if the row does not exist."""
        self.table.delete(**key_dict)
        return None



from operator import itemgetter
import json

from typing import (
    Any,
    List,
    Dict,
    Iterable,
    Optional,
    Set,
    Tuple,
    Union,
)

from cassandra.cluster import ResponseFuture  # type: ignore

from cassio.utils.vector.distance_metrics import distance_metrics

from cassio.table.cql import (
    CQLOpType,
    DELETE_CQL_TEMPLATE,
    SELECT_CQL_TEMPLATE,
    CREATE_INDEX_CQL_TEMPLATE,
    # CREATE_KEYS_INDEX_CQL_TEMPLATE,
    CREATE_ENTRIES_INDEX_CQL_TEMPLATE,
    SELECT_ANN_CQL_TEMPLATE,
)
from cassio.table.table_types import (
    ColumnSpecType,
    RowType,
    RowWithDistanceType,
    normalize_type_desc,
    rearrange_pk_type,
    MetadataIndexingMode,
    MetadataIndexingPolicy,
    is_metadata_field_indexed,
)
from cassio.table.base_table import BaseTable


class BaseTableMixin(BaseTable):
    """All other mixins should inherit from this one."""

    pass


class ClusteredMixin(BaseTableMixin):
    def __init__(
        self,
        *pargs: Any,
        partition_id_type: Union[str, List[str]] = ["TEXT"],
        partition_id: Optional[Any] = None,
        ordering_in_partition: str = "ASC",
        **kwargs: Any,
    ) -> None:
        self.partition_id_type = normalize_type_desc(partition_id_type)
        self.partition_id = partition_id
        self.ordering_in_partition = ordering_in_partition.upper()
        super().__init__(*pargs, **kwargs)

    def _schema_pk(self) -> List[ColumnSpecType]:
        assert len(self.partition_id_type) == 1
        return [
            ("partition_id", self.partition_id_type[0]),
        ]

    def _schema_cc(self) -> List[ColumnSpecType]:
        return self._schema_row_id()

    def _extract_where_clause_blocks(
        self, args_dict: Any
    ) -> Tuple[Any, List[str], Tuple[Any, ...]]:
        """
        If a null partition_id arrives to WHERE construction, it is silently
        discarded from the set of conditions to create.
        This enables e.g. ANN vector search across partitions of a clustered table.

        It is the database's responsibility to raise an error if unacceptable.
        """
        if "partition_id" in args_dict and args_dict["partition_id"] is None:
            cleaned_args_dict = {
                k: v for k, v in args_dict.items() if k != "partition_id"
            }
        else:
            cleaned_args_dict = args_dict
        #
        return super()._extract_where_clause_blocks(cleaned_args_dict)

    def _delete_partition(
        self, is_async: bool, partition_id: Optional[str] = None
    ) -> None:
        _partition_id = self.partition_id if partition_id is None else partition_id
        #
        where_clause = "WHERE " + "partition_id = %s"
        delete_cql_vals = (_partition_id,)
        delete_cql = DELETE_CQL_TEMPLATE.format(
            where_clause=where_clause,
        )
        if is_async:
            return self.execute_cql_async(
                delete_cql, args=delete_cql_vals, op_type=CQLOpType.WRITE
            )
        else:
            self.execute_cql(delete_cql, args=delete_cql_vals, op_type=CQLOpType.WRITE)
            return

    def delete_partition(self, partition_id: Optional[str] = None) -> None:
        self._delete_partition(is_async=False, partition_id=partition_id)
        return None

    def delete_partition_async(
        self, partition_id: Optional[str] = None
    ) -> ResponseFuture:
        return self._delete_partition(is_async=True, partition_id=partition_id)

    def _normalize_kwargs(self, args_dict: Dict[str, Any]) -> Dict[str, Any]:
        # if partition id provided in call, takes precedence over instance value
        arg_pid = args_dict.get("partition_id")
        instance_pid = self.partition_id
        _partition_id = instance_pid if arg_pid is None else arg_pid
        new_args_dict = {
            **{"partition_id": _partition_id},
            **args_dict,
        }
        return super()._normalize_kwargs(new_args_dict)

    def get_partition(
        self, partition_id: Optional[str] = None, n: Optional[int] = None, **kwargs: Any
    ) -> Iterable[RowType]:
        _partition_id = self.partition_id if partition_id is None else partition_id
        get_p_cql_vals: Tuple[Any, ...] = tuple()
        #
        # TODO: work on a columns: Optional[List[str]] = None
        # (but with nuanced handling of the column-magic we have here)
        columns = None
        if columns is None:
            columns_desc = "*"
        else:
            # TODO: handle translations here?
            # columns_desc = ", ".join(columns)
            raise NotImplementedError("Column selection is not implemented.")
        # WHERE can admit other sources (e.g. medata if the corresponding mixin)
        # so we escalate to standard WHERE-creation route and reinject the partition
        n_kwargs = self._normalize_kwargs(
            {
                **{"partition_id": _partition_id},
                **kwargs,
            }
        )
        (args_dict, wc_blocks, wc_vals) = self._extract_where_clause_blocks(n_kwargs)
        # check for exhaustion:
        assert args_dict == {}
        where_clause = "WHERE " + " AND ".join(wc_blocks)
        where_cql_vals = list(wc_vals)
        #
        if n is None:
            limit_clause = ""
            limit_cql_vals = []
        else:
            limit_clause = "LIMIT %s"
            limit_cql_vals = [n]
        #
        select_cql = SELECT_CQL_TEMPLATE.format(
            columns_desc=columns_desc,
            where_clause=where_clause,
            limit_clause=limit_clause,
        )
        get_p_cql_vals = tuple(where_cql_vals + limit_cql_vals)
        return (
            self._normalize_row(raw_row)
            for raw_row in self.execute_cql(
                select_cql,
                args=get_p_cql_vals,
                op_type=CQLOpType.READ,
            )
        )

    def get_partition_async(
        self, partition_id: Optional[str] = None, n: Optional[int] = None, **kwargs: Any
    ) -> ResponseFuture:
        raise NotImplementedError("Asynchronous reads are not supported.")


class MetadataMixin(BaseTableMixin):
    def __init__(
        self,
        *pargs: Any,
        metadata_indexing: Union[Tuple[str, Iterable[str]], str] = "all",
        **kwargs: Any,
    ) -> None:
        self.metadata_indexing_policy = self._normalize_metadata_indexing_policy(
            metadata_indexing
        )
        super().__init__(*pargs, **kwargs)

    @staticmethod
    def _normalize_metadata_indexing_policy(
        metadata_indexing: Union[Tuple[str, Iterable[str]], str]
    ) -> MetadataIndexingPolicy:
        mode: MetadataIndexingMode
        fields: Set[str]
        # metadata indexing policy normalization:
        if isinstance(metadata_indexing, str):
            if metadata_indexing.lower() == "all":
                mode, fields = (MetadataIndexingMode.DEFAULT_TO_SEARCHABLE, set())
            elif metadata_indexing.lower() == "none":
                mode, fields = (MetadataIndexingMode.DEFAULT_TO_UNSEARCHABLE, set())
            else:
                raise ValueError(
                    f"Unsupported metadata_indexing value '{metadata_indexing}'"
                )
        else:
            assert len(metadata_indexing) == 2
            # it's a 2-tuple (mode, fields) still to normalize
            _mode, _field_spec = metadata_indexing
            fields = {_field_spec} if isinstance(_field_spec, str) else set(_field_spec)
            if _mode.lower() in {
                "default_to_unsearchable",
                "allowlist",
                "allow",
                "allow_list",
            }:
                mode = MetadataIndexingMode.DEFAULT_TO_UNSEARCHABLE
            elif _mode.lower() in {
                "default_to_searchable",
                "denylist",
                "deny",
                "deny_list",
            }:
                mode = MetadataIndexingMode.DEFAULT_TO_SEARCHABLE
            else:
                raise ValueError(
                    f"Unsupported metadata indexing mode specification '{_mode}'"
                )
        return (mode, fields)

    def _schema_da(self) -> List[ColumnSpecType]:
        return super()._schema_da() + [
            ("attributes_blob", "TEXT"),
            ("metadata_s", "MAP<TEXT,TEXT>"),
        ]

    def db_setup(self) -> None:
        # Currently: an 'entries' index on the metadata_s column
        super().db_setup()
        #
        entries_index_columns = ["metadata_s"]
        for entries_index_column in entries_index_columns:
            index_name = f"eidx_{entries_index_column}"
            index_column = f"{entries_index_column}"
            create_index_cql = CREATE_ENTRIES_INDEX_CQL_TEMPLATE.format(
                index_name=index_name,
                index_column=index_column,
            )
            self.execute_cql(create_index_cql, op_type=CQLOpType.SCHEMA)
        #
        return

    @staticmethod
    def _serialize_md_dict(md_dict: Dict[str, Any]) -> str:
        return json.dumps(md_dict, separators=(",", ":"), sort_keys=True)

    @staticmethod
    def _deserialize_md_dict(md_string: str) -> Dict[str, Any]:
        return json.loads(md_string)

    @staticmethod
    def _coerce_string(value: Any) -> str:
        if isinstance(value, str):
            return value
        elif isinstance(value, bool):
            # bool MUST come before int in this chain of ifs!
            return json.dumps(value)
        elif isinstance(value, int):
            # we don't want to store '1' and '1.0' differently
            # for the sake of metadata-filtered retrieval:
            return json.dumps(float(value))
        elif isinstance(value, float):
            return json.dumps(value)
        elif value is None:
            return json.dumps(value)
        else:
            # when all else fails ...
            return str(value)

    def _split_metadata_fields(self, md_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Split the *indexed* part of the metadata in separate parts,
        one per Cassandra column.

        Currently: everything gets cast to a string and goes to a single table
        column. This means:
            - strings are fine
            - floats and integers v: they are cast to str(v)
            - booleans: 'true'/'false' (JSON style)
            - None => 'null' (JSON style)
            - anything else v => str(v), no questions asked

        Caveat: one gets strings back when reading metadata
        """

        # TODO: more care about types here
        stringy_part = {k: self._coerce_string(v) for k, v in md_dict.items()}
        return {
            "metadata_s": stringy_part,
        }

    def _normalize_row(self, raw_row: Any) -> Dict[str, Any]:
        md_columns_defaults: Dict[str, Any] = {
            "metadata_s": {},
        }
        pre_normalized = super()._normalize_row(raw_row)
        #
        row_rest = {
            k: v
            for k, v in pre_normalized.items()
            if k not in md_columns_defaults
            if k != "attributes_blob"
        }
        mergee_md_fields = {
            k: v for k, v in pre_normalized.items() if k in md_columns_defaults
        }
        normalized_mergee_md_fields = {
            k: v if v is not None else md_columns_defaults[k]
            for k, v in mergee_md_fields.items()
        }
        r_md_from_s = {
            k: v for k, v in normalized_mergee_md_fields["metadata_s"].items()
        }
        #
        raw_attr_blob = pre_normalized.get("attributes_blob")
        if raw_attr_blob is not None:
            r_attrs = self._deserialize_md_dict(raw_attr_blob)
        else:
            r_attrs = {}
        #
        row_metadata = {
            "metadata": {
                **r_attrs,
                **r_md_from_s,
            },
        }
        #
        normalized = {
            **row_metadata,
            **row_rest,
        }
        return normalized

    def _normalize_kwargs(self, args_dict: Dict[str, Any]) -> Dict[str, Any]:
        _metadata_input_dict = args_dict.get("metadata", {})
        # separate indexed and non-indexed (=attributes) as per indexing policy
        metadata_indexed_dict = {
            k: v
            for k, v in _metadata_input_dict.items()
            if is_metadata_field_indexed(k, self.metadata_indexing_policy)
        }
        attributes_dict = {
            k: self._coerce_string(v)
            for k, v in _metadata_input_dict.items()
            if not is_metadata_field_indexed(k, self.metadata_indexing_policy)
        }
        #
        if attributes_dict != {}:
            attributes_fields = {
                "attributes_blob": self._serialize_md_dict(attributes_dict)
            }
        else:
            attributes_fields = {}
        #
        new_metadata_fields = {
            k: v
            for k, v in self._split_metadata_fields(metadata_indexed_dict).items()
            if v != {} and v != set()
        }
        #
        new_args_dict = {
            **{k: v for k, v in args_dict.items() if k != "metadata"},
            **attributes_fields,
            **new_metadata_fields,
        }
        return super()._normalize_kwargs(new_args_dict)

    def _extract_where_clause_blocks(
        self, args_dict: Any
    ) -> Tuple[Any, List[str], Tuple[Any, ...]]:
        # This always happens after a corresponding _normalize_kwargs,
        # so the metadata, if present, appears as split-fields.
        assert "metadata" not in args_dict
        if "attributes_blob" in args_dict:
            raise ValueError("Non-indexed metadata fields cannot be used in queries.")
        md_keys = {"metadata_s"}
        new_args_dict = {k: v for k, v in args_dict.items() if k not in md_keys}
        # Here the "metadata" entry is made into specific where clauses
        split_metadata = {k: v for k, v in args_dict.items() if k in md_keys}
        these_wc_blocks: List[str] = []
        these_wc_vals_list: List[Any] = []
        # WHERE creation:
        for k, v in sorted(split_metadata.get("metadata_s", {}).items()):
            these_wc_blocks.append(f"metadata_s['{k}'] = %s")
            these_wc_vals_list.append(v)
        # no new kwargs keys are created, all goes to WHERE
        this_args_dict: Dict[str, Any] = {}
        these_wc_vals = tuple(these_wc_vals_list)
        # ready to defer to superclass(es), then collate-and-return
        (s_args_dict, s_wc_blocks, s_wc_vals) = super()._extract_where_clause_blocks(
            new_args_dict
        )
        return (
            {**this_args_dict, **s_args_dict},
            these_wc_blocks + s_wc_blocks,
            tuple(list(these_wc_vals) + list(s_wc_vals)),
        )

    def find_entries(self, n: int, **kwargs: Any) -> Iterable[RowType]:
        columns_desc, where_clause, get_cql_vals = self._parse_select_core_params(
            **kwargs
        )
        limit_clause = "LIMIT %s"
        limit_cql_vals = [n]
        select_vals = tuple(list(get_cql_vals) + limit_cql_vals)
        #
        select_cql = SELECT_CQL_TEMPLATE.format(
            columns_desc=columns_desc,
            where_clause=where_clause,
            limit_clause=limit_clause,
        )
        result_set = self.execute_cql(
            select_cql, args=select_vals, op_type=CQLOpType.READ
        )
        return (self._normalize_row(result) for result in result_set)

    def find_entries_async(self, n: int, **kwargs: Any) -> ResponseFuture:
        raise NotImplementedError("Asynchronous reads are not supported.")


class VectorMixin(BaseTableMixin):
    def __init__(self, *pargs: Any, vector_dimension: int, **kwargs: Any) -> None:
        self.vector_dimension = vector_dimension
        super().__init__(*pargs, **kwargs)

    def _schema_da(self) -> List[ColumnSpecType]:
        return super()._schema_da() + [
            ("vector", f"VECTOR<FLOAT,{self.vector_dimension}>")
        ]

    def db_setup(self) -> None:
        super().db_setup()
        # index on the vector column:
        index_name = "idx_vector"
        index_column = "vector"
        create_index_cql = CREATE_INDEX_CQL_TEMPLATE.format(
            index_name=index_name,
            index_column=index_column,
        )
        self.execute_cql(create_index_cql, op_type=CQLOpType.SCHEMA)
        return

    def ann_search(
        self, vector: List[float], n: int, **kwargs: Any
    ) -> Iterable[RowType]:
        n_kwargs = self._normalize_kwargs(kwargs)
        # TODO: work on a columns: Optional[List[str]] = None
        # (but with nuanced handling of the column-magic we have here)
        columns = None
        if columns is None:
            columns_desc = "*"
        else:
            # TODO: handle translations here?
            # columns_desc = ", ".join(columns)
            raise NotImplementedError("Column selection is not implemented.")
        #
        if all(x == 0 for x in vector):
            # TODO: lift/relax this constraint when non-cosine metrics are there.
            raise ValueError("Cannot use identically-zero vectors in cos/ANN search.")
        #
        vector_column = "vector"
        vector_cql_vals = [vector]
        #
        (
            rest_kwargs,
            where_clause_blocks,
            where_cql_vals,
        ) = self._extract_where_clause_blocks(n_kwargs)
        assert rest_kwargs == {}
        if where_clause_blocks == []:
            where_clause = ""
        else:
            where_clause = "WHERE " + " AND ".join(where_clause_blocks)
        #
        limit_clause = "LIMIT %s"
        limit_cql_vals = [n]
        #
        select_ann_cql = SELECT_ANN_CQL_TEMPLATE.format(
            columns_desc=columns_desc,
            vector_column=vector_column,
            where_clause=where_clause,
            limit_clause=limit_clause,
        )
        #
        select_ann_cql_vals = tuple(
            list(where_cql_vals) + vector_cql_vals + limit_cql_vals
        )
        result_set = self.execute_cql(
            select_ann_cql, args=select_ann_cql_vals, op_type=CQLOpType.READ
        )
        return (self._normalize_row(result) for result in result_set)

    def ann_search_async(
        self, vector: List[float], n: int, **kwargs: Any
    ) -> ResponseFuture:
        raise NotImplementedError("Asynchronous reads are not supported.")

    def metric_ann_search(
        self,
        vector: List[float],
        n: int,
        metric: str,
        metric_threshold: Optional[float] = None,
        **kwargs: Any,
    ) -> Iterable[RowWithDistanceType]:
        rows = list(self.ann_search(vector, n, **kwargs))
        if rows == []:
            return []
        else:
            # sort, cut, validate and prepare for returning
            # evaluate metric
            distance_function, distance_reversed = distance_metrics[metric]
            row_vectors = [row["vector"] for row in rows]
            # enrich with their metric score
            rows_with_metric = list(
                zip(
                    distance_function(row_vectors, vector),
                    rows,
                )
            )
            # sort rows by metric score. First handle metric/threshold
            if metric_threshold is not None:
                if distance_reversed:

                    def _thresholder(mtx, thr):
                        return mtx >= thr

                else:

                    def _thresholder(mtx, thr):
                        return mtx <= thr

            else:
                # no hits are discarded
                def _thresholder(mtx, thr):
                    return True

            #
            sorted_passing_rows = sorted(
                (
                    pair
                    for pair in rows_with_metric
                    if _thresholder(pair[0], metric_threshold)
                ),
                key=itemgetter(0),
                reverse=distance_reversed,
            )
            # return a list of hits with their distance (as JSON)
            enriched_hits = (
                {
                    **hit,
                    **{"distance": distance},
                }
                for distance, hit in sorted_passing_rows
            )
            return enriched_hits

    def metric_ann_search_async(
        self, vector: List[float], n: int, **kwargs: Any
    ) -> ResponseFuture:
        raise NotImplementedError("Asynchronous reads are not supported.")


class ElasticKeyMixin(BaseTableMixin):
    def __init__(self, *pargs: Any, keys: List[str], **kwargs: Any) -> None:
        if "row_id_type" in kwargs:
            raise ValueError("'row_id_type' not allowed for elastic tables.")
        self.keys = keys
        self.key_desc = self._serialize_key_list(self.keys)
        row_id_type = ["TEXT", "TEXT"]
        new_kwargs = {
            **{"row_id_type": row_id_type},
            **kwargs,
        }
        super().__init__(*pargs, **new_kwargs)

    @staticmethod
    def _serialize_key_list(key_vals: List[Any]) -> str:
        return json.dumps(key_vals, separators=(",", ":"), sort_keys=True)

    @staticmethod
    def _deserialize_key_list(keys_str: str) -> List[Any]:
        return json.loads(keys_str)

    def _normalize_row(self, raw_row: Any) -> Dict[str, Any]:
        key_fields = {"key_desc", "key_vals"}
        pre_normalized = super()._normalize_row(raw_row)
        row_key = {k: v for k, v in pre_normalized.items() if k in key_fields}
        row_rest = {k: v for k, v in pre_normalized.items() if k not in key_fields}
        if row_key == {}:
            key_dict = {}
        else:
            # unpack the keys
            assert len(row_key) == 2
            assert self._deserialize_key_list(row_key["key_desc"]) == self.keys
            key_dict = {
                k: v
                for k, v in zip(
                    self.keys,
                    self._deserialize_key_list(row_key["key_vals"]),
                )
            }
        return {
            **key_dict,
            **row_rest,
        }

    def _normalize_kwargs(self, args_dict: Dict[str, Any]) -> Dict[str, Any]:
        # transform provided "keys" into the elastic-representation two-val form
        key_args = {k: v for k, v in args_dict.items() if k in self.keys}
        # the "key" is passed all-or-nothing:
        assert set(key_args.keys()) == set(self.keys) or key_args == {}
        if key_args != {}:
            key_vals = self._serialize_key_list(
                [key_args[key_col] for key_col in self.keys]
            )
            #
            key_args_dict = {
                "key_vals": key_vals,
                "key_desc": self.key_desc,
            }
            other_args_dict = {k: v for k, v in args_dict.items() if k not in self.keys}
            new_args_dict = {
                **key_args_dict,
                **other_args_dict,
            }
        else:
            new_args_dict = args_dict
        return super()._normalize_kwargs(new_args_dict)

    @staticmethod
    def _schema_row_id() -> List[ColumnSpecType]:
        return [
            ("key_desc", "TEXT"),
            ("key_vals", "TEXT"),
        ]


class TypeNormalizerMixin(BaseTableMixin):

    clustered: bool = False
    elastic: bool = False

    def __init__(self, *pargs: Any, **kwargs: Any) -> None:
        if "primary_key_type" in kwargs:
            pk_arg = kwargs["primary_key_type"]
            num_elastic_keys = len(kwargs["keys"]) if self.elastic else None
            col_type_map = rearrange_pk_type(pk_arg, self.clustered, num_elastic_keys)
            new_kwargs = {
                **col_type_map,
                **{k: v for k, v in kwargs.items() if k != "primary_key_type"},
            }
        else:
            new_kwargs = kwargs
        super().__init__(*pargs, **new_kwargs)



from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple, Union

ColumnSpecType = Tuple[str, str]
RowType = Dict[str, Any]
RowWithDistanceType = Dict[str, Any]
SessionType = Any


class MetadataIndexingMode(Enum):
    DEFAULT_TO_UNSEARCHABLE = 1
    DEFAULT_TO_SEARCHABLE = 2


MetadataIndexingPolicy = Tuple[MetadataIndexingMode, Set[str]]


def is_metadata_field_indexed(field_name: str, policy: MetadataIndexingPolicy) -> bool:
    p_mode, p_fields = policy
    if p_mode == MetadataIndexingMode.DEFAULT_TO_UNSEARCHABLE:
        return field_name in p_fields
    elif p_mode == MetadataIndexingMode.DEFAULT_TO_SEARCHABLE:
        return field_name not in p_fields
    else:
        raise ValueError(f"Unexpected metadata indexing mode {p_mode}")


def normalize_type_desc(type_desc: Union[str, List[str]]) -> List[str]:
    if isinstance(type_desc, str):
        return [type_desc]
    else:
        return type_desc


def rearrange_pk_type(
    pk_type: Union[str, List[str]],
    clustered: bool = False,
    num_elastic_keys: Optional[int] = None,
) -> Dict[str, List[str]]:
    """A compatibility layer with the 'primary_key_type' specifier on init."""
    _pk_type = normalize_type_desc(pk_type)
    if clustered:
        pk_type, rest_type = _pk_type[0:1], _pk_type[1:]
        if num_elastic_keys:
            assert len(rest_type) == num_elastic_keys
            return {
                "partition_id_type": pk_type,
            }
        else:
            return {
                "row_id_type": rest_type,
                "partition_id_type": pk_type,
            }
    else:
        if num_elastic_keys:
            assert len(_pk_type) == num_elastic_keys
            return {}
        else:
            return {
                "row_id_type": _pk_type,
            }



from typing import List, Dict, Tuple, Callable

import numpy as np


# distance definitions. These all work batched in the first argument.
def distance_dot_product(
    embedding_vectors: List[List[float]], reference_embedding_vector: List[float]
) -> List[float]:
    """
    Given a list [emb_i] and a reference rEmb vector,
    return a list [distance_i] where each distance is
        distance_i = distance(emb_i, rEmb)
    At the moment only the dot product is supported
    (which for unitary vectors is the cosine difference).

    Not particularly optimized.
    """
    v1s = np.array(embedding_vectors, dtype=float)
    v2 = np.array(reference_embedding_vector, dtype=float)
    return list(
        np.dot(
            v1s,
            v2.T,
        )
    )


def distance_cos_difference(
    embedding_vectors: List[List[float]], reference_embedding_vector: List[float]
) -> List[float]:
    v1s = np.array(embedding_vectors, dtype=float)
    v2 = np.array(reference_embedding_vector, dtype=float)
    return list(
        np.dot(
            v1s,
            v2.T,
        )
        / (np.linalg.norm(v1s, axis=1) * np.linalg.norm(v2))
    )


def distance_l1(
    embedding_vectors: List[List[float]], reference_embedding_vector: List[float]
) -> List[float]:
    v1s = np.array(embedding_vectors, dtype=float)
    v2 = np.array(reference_embedding_vector, dtype=float)
    return list(np.linalg.norm(v1s - v2, axis=1, ord=1))


def distance_l2(
    embedding_vectors: List[List[float]], reference_embedding_vector: List[float]
) -> List[float]:
    v1s = np.array(embedding_vectors, dtype=float)
    v2 = np.array(reference_embedding_vector, dtype=float)
    return list(np.linalg.norm(v1s - v2, axis=1, ord=2))


def distance_max(
    embedding_vectors: List[List[float]], reference_embedding_vector: List[float]
) -> List[float]:
    v1s = np.array(embedding_vectors, dtype=float)
    v2 = np.array(reference_embedding_vector, dtype=float)
    return list(np.linalg.norm(v1s - v2, axis=1, ord=np.inf))


# The tuple is:
#   (
#       function,
#       sorting 'reverse' argument, nearest-to-farthest
#   )
# (i.e. True means that:
#     - in that metric higher is closer and that
#     - cutoff should be metric > threshold)
distance_metrics: Dict[str, Tuple[Callable, bool]] = {
    "cos": (
        distance_cos_difference,
        True,
    ),
    "dot": (
        distance_dot_product,
        True,
    ),
    "l1": (
        distance_l1,
        False,
    ),
    "l2": (
        distance_l2,
        False,
    ),
    "max": (
        distance_max,
        False,
    ),
}



"""
Compatibility layer for legacy VectorTable (used by LangChain integration
(as of August 2023).

Note: This is to be replaced by direct usage of the table-class-hierarchy classes.
"""

from warnings import warn
from typing import List, Dict, Any, Optional

from cassandra.cluster import ResponseFuture  # type: ignore

from cassio.table.table_types import RowType
from cassio.table.tables import (
    MetadataVectorCassandraTable,
)

new_columns_to_legacy = {
    "row_id": "document_id",
    "body_blob": "document",
    "vector": "embedding_vector",
}
legacy_columns_to_new = {v: k for k, v in new_columns_to_legacy.items()}


class VectorTable:
    """
    This class is a rewriting of the VectorTable created for use in LangChain
    integration, this time relying on the class-table-hierarchy (cassio.table.*).

    It mostly provides a translation layer between parameters and key names,
    using a metadata+vector table class internally.

    Additional kwargs, for use in this new table class, are passed as they are
    in order to enable their usage already before adapting the LangChain
    integration code.
    """

    DEPRECATION_MESSAGE = (
        "Class `VectorTable` is a legacy construct and "
        "will be deprecated in future versions of CassIO."
    )

    def __init__(self, *pargs: Any, **kwargs: Dict[str, Any]):
        #
        warn(self.DEPRECATION_MESSAGE, DeprecationWarning, stacklevel=2)
        #
        if "embedding_dimension" in kwargs:
            vector_dimension = kwargs["embedding_dimension"]
            new_kwargs = {
                **{
                    k: v
                    for k, v in kwargs.items()
                    if k != "embedding_dimension"
                    # let's get rid of the infamous 'auto_id' here:
                    if k != "auto_id"
                },
                **{"vector_dimension": vector_dimension},
            }
        else:
            new_kwargs = kwargs
        # this legacy VectorTable will have everything indexed for search:
        md_kwargs = {
            **{"metadata_indexing": "all"},
            **new_kwargs,
        }
        #
        self.table = MetadataVectorCassandraTable(*pargs, **md_kwargs)

    def search(
        self,
        embedding_vector: List[float],
        top_k: int,
        metric: str = "cos",
        metric_threshold: Optional[float] = None,
        **kwargs: Any,
    ) -> List[RowType]:
        # get rows by ANN
        enriched_hits = self.table.metric_ann_search(
            vector=embedding_vector,
            n=top_k,
            metric=metric,
            metric_threshold=metric_threshold,
            **kwargs,
        )
        #
        return [self._make_dict_legacy(rich_hit) for rich_hit in enriched_hits]

    def put(
        self,
        document: str,
        embedding_vector: List[float],
        document_id: Any,
        metadata: Dict[str, Any] = {},
        ttl_seconds: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        self.table.put(
            row_id=document_id,
            body_blob=document,
            vector=embedding_vector,
            metadata=metadata or {},
            ttl_seconds=ttl_seconds,
            **kwargs,
        )

    def put_async(
        self,
        document: str,
        embedding_vector: List[float],
        document_id: Any,
        metadata: Dict[str, Any],
        ttl_seconds: int,
        **kwargs: Any,
    ) -> ResponseFuture:
        return self.table.put_async(
            row_id=document_id,
            body_blob=document,
            vector=embedding_vector,
            metadata=metadata or {},
            ttl_seconds=ttl_seconds,
            **kwargs,
        )

    def get(self, document_id: Any, **kwargs: Any) -> Optional[RowType]:
        row_or_none = self.table.get(row_id=document_id, **kwargs)
        if row_or_none:
            return self._make_dict_legacy(row_or_none)
        else:
            return row_or_none

    def delete(self, document_id: Any, **kwargs: Any) -> None:
        self.table.delete(row_id=document_id, **kwargs)
        return None

    def clear(self) -> None:
        self.table.clear()
        return None

    @staticmethod
    def _make_dict_legacy(new_dict: RowType) -> RowType:
        return {new_columns_to_legacy.get(k, k): v for k, v in new_dict.items()}
